
--
--  TjOptions.lua
--    by Tuhljin
--
--  Library for easy creation and management of Interface Options panels.
--
--  === PANEL CREATION API ===
--
--  panel = TjOptions.CreatePanel(name, parent, [isDefaultInGroup,] tab)
--    Creates a new Interface Options panel. The contents of the panel will not be created immediately, instead
--    being dynamically created if and when it is necessary.
--    Arguments:
--      name     String. The panel's name, being the text that is used to indicate this panel in the Interface
--               Options frame.
--      parent   String or nil. The name of the panel which will be the new panel's parent, or nil for no parent.
--      isDefaultInGroup
--               Any. If it evaluates to true and a parent created by TjOptions.CreatePanelGroup() is specified, this
--               panel will be shown when the parent panel is selected.
--      tab      Table. The panel's contents, using the specifications indicated in the "PANEL CONTENTS TABLE
--               SPECIFICATIONS" section below.
--    Returns:
--      panel    Table. The new panel (frame).
--
--  panel = TjOptions.CreatePanelGroup(name, parent, [isDefaultInGroup[, sound]])
--    Creates a new Interface Options panel that acts solely as a group header for other panels. If this panel is
--    selected, one of its children (if it has any) will be displayed.
--    Arguments:
--      name, parent, isDefaultInGroup
--               As TjOptions.CreatePanel()'s arguments.
--      sound    String or nil. A sound file to play when the panel is selected and one of its children is shown. Use
--               nil to play no sound, or an empty string ("") to use the suggested sound ("igCharacterInfoTab").
--    Returns:
--      panel    Table. The new panel (frame).
--
--  === POST-CREATION API ===
--
--  In the following, myPanelGroup is the table returned by a TjOptions.CreatePanelGroup() call while myPanel is that
--  returned by TjOptions.CreatePanel().
--
--  myPanelGroup:SetDefaultPanel("PanelName")  -OR-
--  myPanelGroup:SetDefaultPanel(PanelFrame)
--    Change the panel which will be shown when myPanelGroup is selected. The argument is either a string giving the
--    name of the panel or the panel frame itself.
--
--  myPanelGroup:SetSound("SoundFile")
--    Change the sound file associated with this panel, as per the sound argument of TjOptions.CreatePanelGroup().
--
--  myPanel:LoadVariables()
--    Cause all items on the panel to reflect the saved variables they represent (checking/unchecking appropriate
--    checkboxes, etc.). This is called automatically when the panel is shown or set to default settings.
--
--  changesInProgress, canceling, defaulting, loading = myPanel:IsChangeInProgress()
--    Returns true if the panel is currently in the process of adjusting its items' settings, or nil otherwise.
--    Also returns 3 other values, true or nil, based on which changes are under way:
--      canceling    True if panel is reverting to previous settings, as occurs when the player hits the Cancel
--                   button to close the Interface Options frame (if myPanel was shown since the frame was opened).
--      defaulting   True if panel is reverting to default settings.
--      loading      True if panel is in the middle of a LoadVariables() call.
--   This function can be useful to check in an OnChange function (see table specifications, below) if it should
--   react differently in this case. Example usage in an item's OnChange function:
--      local function ItemChanged(self, key, val)
--        if (myPanel:IsChangeInProgress()) then  return;  end
--        -- ... other stuff here
--      end
--
--
--  === PANEL CONTENTS TABLE SPECIFICATIONS ===
--
--  The table used as the fourth (or third, if isDefaultInGroup is omitted) argument of a TjOptions.CreatePanel()
--  call should use the following format. In the following, myPanel is the table returned by a
--  TjOptions.CreatePanel() call.
--
--  Baseline Fields:
--    All fields are optional except "items."
--    title         String. Text of the large label used at the top of the panel. (The resulting label can be
--                  accessed at myPanel.titleLabel.)
--    items         Table. The items, such as checkboxes, used on this panel. See "Item Fields" below.
--    itemspacing   Integer. The size of the empty buffer space between items. A default value is used if nil.
--    variables     Table. The variables that will be altered when the player manipulates the panel's items. See
--                  "Variables Table" below. It is expected that any given table will be used for only one panel,
--                  and only variables to be handled by TjOptions are contained therein; unexpected results may occur
--                  otherwise.
--    defaults      Table. Default settings for the variables. This should be a duplicate of the "variables" table
--                  except its values are what you want variables to be set to when the player sets this panels'
--                  settings to the default.
--    OnChange      Function. The given function will be called when TjOptions may have changed the value of any of
--                  the panel's items' saved variables. It is called once for each item that may have changed and is
--                  passed the item's frame as the first argument, the key of the associated variable (i.e.
--                  "DoSomething") as the second one, and the value of that variable as the third. Also see the
--                  OnChange item field, below.
--    OnShow        Function. Called when the panel is shown, with the panel as its first argument. The function is
--                  called after TjOptions handles its standard OnShow functionality, including LoadVariables().
--    OnDefault     Function. Called when the panel's options are set to the default.
--    OnBuild       Function. Called when the panel's contents are built, with the panel as its first argument. This
--                  is called only once, either when the panel is shown for the first time (before the given OnShow
--                  function is called) or when the player sets all interface options to the default. At the time of
--                  this call, the panel's contents (checkboxes, etc.) have been created but LoadVariables() hasn't
--                  been called so their settings may be wrong. LoadVariables() is called after the OnBuild function,
--                  so you could change the values of your variables at this time so the new values will be used
--                  (except they will still be set to their defaults, if any, if that's what triggered this).
--    scrolling     Any. If it evaluates to true, the panel's items will be placed inside a scolling frame. If this
--                  is a string, it will be the scroll frame's global variable name.
--
--  Item Fields - general (for all items):
--    All general item fields are optional.
--    type          String. The type of item. Can be:
--                    "checkbox"       A checkbox. This is the default if type isn't specified.
--                    "checkboxsmall"  A smaller checkbox with text that is normally yellow.
--                    "dropdown"       A dropdown menu. Requires the TjDropDownMenu library (a separate Lua file).
--                    "label"          A white text label.
--                    "labelnorm"      A text label using the normal WoW font color.
--    variable      String. The key of the variable this item will manipulate. For example, if the type is "checkbox"
--                  and this field is "DoSomething," when the checkbox is clicked, the "DoSomething" entry in the
--                  "variables" table (from the baseline field) is set to true if checked or false otherwise.
--    text          String. Text of the label displayed alongside the item.
--    name          String. The name of the global variable to give the new object. Use this if your addon needs
--                  access to the object, but remember that it won't actually be created until it is needed.
--    tooltip       String or function. If a string, tooltip text to display. If a function, it is called after
--                  the tooltip's owner is set and is ready to have text added to it. The item's frame is passed as
--                  the first argument and the tooltip object is passed as the second argument. Your function doesn't
--                  need to show the tooltip ("GameTooltip:Show()") since that will be done for you by TjOptions.
--    tooltip2      String. Additional tooltip text (uses a smaller font). Used only if "tooltip" field is a string.
--    xOffset       Integer. Offset the x position of the item by this amount. Items following it will also be offset
--                  by this amount until another is specified. (Set to 0 to end previous offsets.)
--    topBuffer     Integer. Additional space to add between this item and the previous one, or fewer if negative.
--    btmBuffer     Integer. Additional space to add between this item and the next one, or fewer if negative.
--    OnChange      Function. If this is defined, it will override the panel's OnChange field where this specific
--                  item is concerned. If desired, you can set this to something other than a function (except for
--                  nil or false) to prevent the panel's OnChange function from being called without calling a
--                  different function.
--
--  Item Fields - label or labelnorm:
--    Two additional optional fields are available for labels.
--    width         Integer. The width of the label. Use this to allow automatic wordwrapping.
--    justifyH      String. The horizontal justification of the text (such as "LEFT"). Labels are center-justified by
--                  default.
--
--  Item Fields - dropdown:
--    Three additional fields are available for dropdown menu items. The menu field is required.
--    menu          Table. Used as the third argument in TjDropDownMenu.CreateDropDown(). See TjDropDownMenu.lua.
--                  If you use the tooltip item fields, they will be used instead of those found in this table when
--                  the cursor is over the dropdown frame. Set tooltip to 0 show no tooltip. (The menu table's fields
--                  will still be used for the dropped-down list itself.)
--    width         Integer. The width of the primary portion of the dropdown frame.
--    displayMode   String. The display mode of the dropdown menu. Usually left nil or set to "MENU" (w/o quotes).
--
--  Variables Table: This table is intended to be used by your addon to determine the addon's options/configuration.
--  It might be the same table you declare as a saved variable in the addon's TOC file. The keys used should
--  correspond to those indicated by the "variable" field of the entries in the "items" table.
--
--  A simple example function call using such a table follows:
--
--  local panel = TjOptions.CreatePanel("My AddOn", nil, {
--  	title = "My Addon",
--  	items = {
--  		 { type = "checkbox", variable = "Enabled", text = "Enable Stuff" },
--    		 { type = "checkboxsmall", variable = "VerboseMode", text = "Verbose Mode",
--    		   tooltip = "Increase verbosity", tooltip2 = "Makes things excessively wordy."  }
--  	},
--  	variables = MyAddon_Settings,
--  	defaults = MyAddon_Settings_Default
--  });
--
--

local THIS_VERSION = 0.23

if (not TjOptions or TjOptions.Version < THIS_VERSION) then
  TjOptions = {};
  TjOptions.Version = THIS_VERSION;

  local defaultSound = "igCharacterInfoTab";

  local Panels = {};

  local ItemGlobalName = "TjOptionsItemNumber"  -- Used to give unique global names to items that don't specify one.
  local ScrollFrameGlobalName = "TjOptionsScrollFrame"
  local numItems, numScrollFrames = 0, 0

  local function copytab(from, to)
    for k,v in pairs(from) do
      if(type(v) == "table") then
        to[k] = {}
        copytab(v, to[k]);
      else
        to[k] = v;
      end
    end
  end

  local function GetPanel(obj)
    local panel = obj:GetParent()
    -- If scrolling frame is used, the main panel frame is a few levels up from the object:
    if (panel.TjOpt_isScrollChild) then  return panel:GetParent():GetParent();  end
    return panel;
  end

  local function GetItems(panel)
    if (panel.TjOpt_scrollchild) then  return panel.TjOpt_scrollchild:GetChildren();  end
    return panel:GetChildren();
  end


-- PANEL ITEM INTERACTION
----------------------------

  local function ItemOnEnter(self)
    local tip = self.TjOpt_tab.tooltip
    if (tip) then
      GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
      GameTooltip:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
      if (type(tip) == "string") then
        GameTooltip:AddLine(tip, nil, nil, nil, 1);
        if (self.TjOpt_tab.tooltip2) then
         GameTooltip:AddLine(self.TjOpt_tab.tooltip2, nil, nil, nil, 1)
        end
      elseif (type(tip) == "function") then
        tip(self, GameTooltip)
      end
      GameTooltip:Show()
    end
  end

  local function ItemOnLeave(self)
    GameTooltip:Hide();
  end

  local function sendOnChange(obj, key, val)
    local func = obj.TjOpt_tab.OnChange or GetPanel(obj).TjOpt_tab.OnChange
    if (type(func) == "function") then
      func(obj, key, val)
    end
  end

  local function sendOnChange_recursive(vars, obj, ...)
    if (obj) then
      if (obj.TjOpt_tab) then
        local key = obj.TjOpt_tab.variable
        sendOnChange(obj, key, vars[key])
      end
      sendOnChange_recursive(vars, ...)
    end
  end

  local function LoadVar(obj, vars)
    if (obj.TjOpt_tab and obj.TjOpt_tab.variable) then
      local v = vars[obj.TjOpt_tab.variable]
      local t = obj.TjOpt_tab.type
      if (t == "checkbox" or t == "checkboxsmall") then
        if (v) then
          obj:SetChecked(1);
        else
          obj:SetChecked(0);
          -- Prevent nil value from causing inconsistencies when copytab function is used:
          if (v == nil) then  vars[obj.TjOpt_tab.variable] = false;  end
        end
      elseif (t == "dropdown") then
        if (obj:GetSelectedValue() ~= v) then  -- Needed to prevent DropdownOnSelect call.
          obj:SetSelectedValue(v)
        end
      else
        error("Unknown type: "..(t or "nil"))
      end
    end
  end

  local function LoadVars_recursive(vars, obj, ...)
    if (obj) then
      LoadVar(obj, vars)
      LoadVars_recursive(vars, ...)
    end
  end

  local function SaveVar(obj, val)
    if (obj.TjOpt_tab and obj.TjOpt_tab.variable) then
      local vars = GetPanel(obj).TjOpt_tab.variables
      if (vars) then
        local v
        local t = obj.TjOpt_tab.type
        if (t == "checkbox" or t == "checkboxsmall") then
          if (obj:GetChecked()) then  v = true;  else  v = false;  end
        elseif (t == "dropdown") then
          v = val
        else
          error("Unknown type: "..(t or "nil"))
        end
        vars[obj.TjOpt_tab.variable] = v
        return obj.TjOpt_tab.variable, v;
      end
    end
  end

  local function CheckboxOnClick(self)
    PlaySound("igMainMenuOptionCheckBoxOn")
    local key, val = SaveVar(self)
    sendOnChange(self, key, val)
  end

  local function DropdownOnSelect(self, val, tab)
    local key
    key, val = SaveVar(self, val)
    sendOnChange(self, key, val)
  end


-- BUILD PANEL CONTENTS
--------------------------

  local function CreateCheckbox(globalname, parent, text, template)
    template = template or "InterfaceOptionsCheckButtonTemplate"
    local frame = CreateFrame("CheckButton", globalname, parent, template)
    frame:SetScript("OnClick", CheckboxOnClick)
    frame:SetScript("OnEnter", ItemOnEnter)
    frame:SetScript("OnLeave", ItemOnLeave)
    if (text) then  getglobal(globalname.."Text"):SetText(text);  end
    return frame;
  end

  local function CreateLabel(globalname, parent, text, width, justifyH, font)
    font = font or "GameFontHighlight"
    local label = parent:CreateFontString(globalname, "ARTWORK", font);
    if (width) then  label:SetWidth(width);  end
    if (justifyH) then  label:SetJustifyH(justifyH);  end
    label:SetText(text);
    return label;
  end

  local function BuildPanelContents(panel)
    panel.TjOpt_built = true;
    local tab = panel.TjOpt_tab
    if (tab.title) then
      panel.titleLabel = panel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge");
      panel.titleLabel:SetPoint("TOPLEFT", panel, "TOPLEFT", 16, -16);
      panel.titleLabel:SetText(tab.title);
    end

    local workpanel
    if (tab.scrolling) then
      local scrollname = type(tab.scrolling) == "string" and tab.scrolling or ScrollFrameGlobalName..(numScrollFrames + 1)
      local scrollframe = CreateFrame("ScrollFrame", scrollname, panel, "UIPanelScrollFrameTemplate")
      numScrollFrames = numScrollFrames + 1
      scrollframe:SetWidth(382)
      if (panel.titleLabel) then
        scrollframe:SetHeight(398 - panel.titleLabel:GetHeight())
        scrollframe:SetPoint("TOPLEFT", panel.titleLabel, "TOPLEFT", -16, -21)
      else
        scrollframe:SetHeight(410)
        scrollframe:SetPoint("TOPLEFT", panel, "TOPLEFT", 0, -9)
      end
      local scrollchild = CreateFrame("Frame", nil, scrollframe)
      scrollchild:SetHeight(370)
      scrollchild:SetWidth(370)
      scrollchild:SetPoint("TOPLEFT", scrollframe, "TOPLEFT")
      --scrollchild:Show(); scrollframe:Show();
      scrollframe:SetScrollChild(scrollchild)
      panel.TjOpt_scrollchild = scrollchild
      scrollchild.TjOpt_isScrollChild = true
      workpanel = scrollchild
    else
      workpanel = panel
    end

    local y = tab.itemspacing or 6; y = -y
    local obj, relativeObj, t, n, offx, offy, correctx
    local nextoffx, nextoffy = 0, 0
    local xOffset_total = 0
    for i, v in ipairs(tab.items) do
      correctx, offx, offy = nextoffx, 0, nextoffy - (v.topBuffer or 0)
      nextoffx, nextoffy = 0, (v.btmBuffer or 0) * -1
      t = v.type
      if (not t) then  v.type, t = "checkbox", "checkbox";  end
      n = v.name or ItemGlobalName..(numItems + 1)
      if (t == "checkbox") then
        obj = CreateCheckbox(n, workpanel, v.text)
      elseif (t == "checkboxsmall") then
        obj = CreateCheckbox(n, workpanel, v.text, "InterfaceOptionsSmallCheckButtonTemplate")
      elseif (t == "dropdown") then
        assert(TjDropDownMenu, "TjDropDownMenu library not found.");
        assert(v.menu, "Required item field missing: menu.");
        obj = TjDropDownMenu.CreateDropDown(n, workpanel, v.menu, nil, nil, v.displayMode, 5);
        obj:SetDropDownWidth(v.width or 85)
        obj:OnSelect(DropdownOnSelect)
        obj.tooltip, obj.tooltip2 = v.tooltip, v.tooltip2
        if (v.text) then
          obj:SetLabel(v.text)
          offy = offy - 17
        else
          offy = offy - 4
        end
        offx = offx - 13
      elseif (t == "label") then
        obj = CreateLabel(n, workpanel, v.text, v.width, v.justifyH)
        offy = offy - 2
        nextoffy = nextoffy - 2
      elseif (t == "labelnorm") then
        obj = CreateLabel(n, workpanel, v.text, v.width, v.justifyH, "GameFontNormal")
        offy = offy - 2
        nextoffy = nextoffy - 2
      end
      if (obj) then
        numItems = numItems + 1
        obj.TjOpt_tab = v
        local x = offx + correctx
        if (v.xOffset) then
          x = x - xOffset_total + v.xOffset
          xOffset_total = v.xOffset
        end
        if (relativeObj) then
          obj:SetPoint("TOPLEFT", relativeObj, "BOTTOMLEFT", x, y + offy)
        else
          if (tab.scrolling) then
            obj:SetPoint("TOPLEFT", workpanel, "TOPLEFT", x + 14, offy - 5)
          elseif (panel.titleLabel) then
            obj:SetPoint("TOPLEFT", panel.titleLabel, "BOTTOMLEFT", x - 2, offy - 12)
          else
            obj:SetPoint("TOPLEFT", workpanel, "TOPLEFT", x + 14, offy - 16)
          end
        end
        relativeObj = obj;
        nextoffx = nextoffx - offx;
      end
      obj = nil;
    end

    local func = tab.OnBuild
    if (type(func) == "function") then
      func(panel)
    end
    if (tab.variables) then
      tab.cancelTo = {}
      copytab(tab.variables, tab.cancelTo);  -- Remember current settings in case Cancel is pressed.
    end
  end


-- PANEL USE FUNCTIONS
-------------------------

  local function OpenGroup(self)
    local panel = self.TjOpt_panel
    if (type(panel) == "string") then  panel = Panels[panel];  end
    if (type(panel) == "table") then
      if (self.TjOpt_sound) then  PlaySound(self.TjOpt_sound);  end
      InterfaceOptionsFrame_OpenToCategory(panel)
    end
  end

  local function funcOnShow(self)
    if (not self.TjOpt_built) then  BuildPanelContents(self);  end
    self:LoadVariables()
    self.TjOpt_possiblechange = true
    local func = self.TjOpt_tab.OnShow
    if (type(func) == "function") then
      func(self)
    end
  end

  local function funcOkay(self)
    if (self.TjOpt_built and self.TjOpt_tab.variables) then
      -- Remember current settings in case Cancel is pressed.
      copytab(self.TjOpt_tab.variables, self.TjOpt_tab.cancelTo);
    end
    self.TjOpt_possiblechange = nil
  end

  local function funcCancel(self)
    if (self.TjOpt_tab.cancelTo and self.TjOpt_possiblechange) then
      self.TjOpt_tab.isCanceling = true
      -- Revert to previous settings.
      copytab(self.TjOpt_tab.cancelTo, self.TjOpt_tab.variables);
      sendOnChange_recursive(self.TjOpt_tab.variables, GetItems(self))
      self.TjOpt_tab.isCanceling = nil
    end
    self.TjOpt_possiblechange = nil
  end

  local function funcDefault(self)
    if (self.TjOpt_tab.defaults) then
      self.TjOpt_tab.isDefaulting = true
      if (not self.TjOpt_built) then  BuildPanelContents(self);  end
      copytab(self.TjOpt_tab.defaults, self.TjOpt_tab.variables);
      self:LoadVariables()  -- This has switched between being necessary and being redundant (due to OnShow being
      -- called automatically) multiple times as the beta has progressed. Leaving it in for now; test it after WotLK
      -- has been released for a while.
      self.TjOpt_possiblechange = true
      sendOnChange_recursive(self.TjOpt_tab.variables, GetItems(self))
      self.TjOpt_tab.isDefaulting = nil
      if (not InterfaceOptionsFrame:IsShown()) then
      -- Remember current settings if Interface Options frame isn't shown.
      -- (Allow Cancel to revert this change if it is shown.)
        copytab(self.TjOpt_tab.variables, self.TjOpt_tab.cancelTo);
      end
    end
    local func = self.TjOpt_tab.OnDefault
    if (type(func) == "function") then
      func(self)
    end
  end


-- PANEL API
---------------

  local function LoadVariables(self)
    if (self.TjOpt_built and self.TjOpt_tab.variables) then
      self.TjOpt_tab.isLoading = true
      LoadVars_recursive(self.TjOpt_tab.variables, GetItems(self))
      self.TjOpt_tab.isLoading = nil
    end
  end

-- Call in one of two ways:  myPanelGroup:SetDefaultPanel("PanelName")  -OR-
--                           myPanelGroup:SetDefaultPanel(PanelFrame)
  local function SetDefaultPanel(self, panel)
    if (type(panel) ~= "string" and type(panel) ~= "table") then
      error("SetDefaultPanel(): Invalid argument. Expected string or table; got "..type(panel)..".")
    end
    self.TjOpt_panel = panel
  end

-- Call like so:  myPanelGroup:SetSound("SoundFile")
-- Use nil as the arg for no sound. Use an empty string ("") to use the default sound.
  local function SetSound(self, sound)
    if (type(sound) ~= "nil" and type(sound) ~= "string") then
      error("SetSound(): Invalid argument. Expected string or nil; got "..type(sound)..".")
    end
    if (sound == "") then  sound = defaultSound;  end
    self.TjOpt_sound = sound
  end

  local function IsChangeInProgress(self)
    local c, d, l = self.TjOpt_tab.isCanceling, self.TjOpt_tab.isDefaulting, self.TjOpt_tab.isLoading
    return (c or d or l), c, d, l;
  end


-- PANEL CREATION API
------------------------

  function TjOptions.CreatePanel(name, parent, isDefaultInGroup, tab)
    if (type(tab) == "nil" and type(isDefaultInGroup) == "table") then
      tab = isDefaultInGroup;
      isDefaultInGroup = nil;
    end
    if (type(tab) ~= "table" or type(name) ~= "string" or (type(parent) ~= "nil" and type(parent) ~= "string")) then
      error("Invalid arg(s). Usage: TjOptions.CreatePanel(stringName, stringParentPanelName/nil, [isDefaultInGroup (boolean/any),] tablePanelLayout)")
    elseif (Panels[name]) then
      error("TjOptions.CreatePanel(): A panel named '"..name.."' already exists.")
    elseif (isDefaultInGroup and parent and not Panels[parent]) then
      error("TjOptions.CreatePanel(): Cannot set as default in group because given parent was not found in TjOP data.")
    elseif (type(tab.items) ~= "table" or not tab.items[1]) then
      error("TjOptions.CreatePanel(): Table invalid - 'items' key missing or invalid.")
    end

    local panel = CreateFrame("Frame")
    panel.name = name
    panel.parent = parent
    panel.okay = funcOkay
    panel.cancel = funcCancel
    panel.default = funcDefault
    panel.TjOpt_tab = tab
    panel.LoadVariables = LoadVariables
    panel.IsChangeInProgress = IsChangeInProgress
    panel:SetScript("OnShow", funcOnShow)
    if (parent and (isDefaultInGroup or not Panels[parent].TjOpt_panel)) then
      Panels[parent].TjOpt_panel = panel;
    end
    Panels[name] = panel
    InterfaceOptions_AddCategory(panel)
    return panel;
  end

  function TjOptions.CreatePanelGroup(name, parent, isDefaultInGroup, sound)
    if (type(name) ~= "string" or (type(parent) ~= "nil" and type(parent) ~= "string") or (type(sound) ~= "nil" and type(sound) ~= "string")) then
      error("Invalid arg(s). Usage: TjOptions.CreatePanelGroup(stringName, stringParentPanelName/nil, [isDefaultInGroup (boolean/any)[, stringSoundFile]])")
    elseif (Panels[name]) then
      error("TjOptions.CreatePanel(): A panel named '"..name.."' already exists.")
    elseif (isDefaultInGroup and parent and not Panels[parent]) then
      error("TjOptions.CreatePanelGroup(): Cannot set as default in group because given parent was not found in TjOP data.")
    end

    local panel = CreateFrame("Frame")
    panel.name = name
    panel.parent = parent
    panel.SetDefaultPanel = SetDefaultPanel
    panel.SetSound = SetSound
    panel:SetSound(sound)
    panel:SetScript("OnShow", OpenGroup)
    if (parent and (isDefaultInGroup or not Panels[parent].TjOpt_panel)) then
      Panels[parent].TjOpt_panel = panel;
    end
    Panels[name] = panel
    InterfaceOptions_AddCategory(panel)
    return panel;
  end

end
